// Code based on Sivan Toledo's original published at http://www.tau.ac.il/~stoledo/lego/AVR-ir/
// Modified by Jonathan Oakley

/*
 * I2C routines
 */
#include <stdlib.h>
#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>
//#include <avr/signal.h>
#include <avr/pgmspace.h>
#include "uart.h"
#include "PFir.h"

#define I2C_MAX_MSG 16
static volatile uint8_t i2c_in_buf[I2C_MAX_MSG]; // max buffer as defined by Lego 
//static volatile uint8_t i2c_dbg_buf[I2C_MAX_MSG]; // max buffer as defined by Lego 
static volatile uint8_t* i2c_out_msg;
volatile PFmsg i2c_out_msgA[PF_MSG_SLOTS],i2c_out_buf; // 4 channels, but can address 16 outputs
volatile PFmsg i2c_snd_msgA[PF_MSG_SLOTS]; // buffers for messages to send

static volatile uint8_t i2c_in_msg_len; //,dbg_idx=0;
static volatile uint8_t i2c_out_msg_len=0;
static volatile uint8_t i2c_proc_cnt=0;

// Need to put x46 in your code, though...
#define SLAVE_ADDRESS 0x23

volatile uint8_t i2c_processed = 0;
static volatile uint8_t i2c_new_input = 0;

void i2c_process_in_msg()
{
	// i2c_proc_cnt++;
	if (!i2c_processed) // don't process twice!
		switch (i2c_in_buf[0])
		{
			case 0x01:
			/*	pf_repeats = 5; // one shot
				pf_cmd = i2c_in_buf[1];
				pf_cmd <<= 8;
				pf_cmd |= i2c_in_buf[2];
				pf_send_start(); */  
				pf_send(((uint16_t) i2c_in_buf[1])<<8 | i2c_in_buf[2],5);
				break;

			case 0x02:
			/*	pf_repeats = -1; // one shot
				pf_cmd = i2c_in_buf[1];
				pf_cmd <<= 8;
				pf_cmd |= i2c_in_buf[2];
				pf_send_start(); */
				pf_send(((uint16_t) i2c_in_buf[1])<<8 | i2c_in_buf[2],-1);
				break;

			case 0x40:
				i2c_out_buf.PFstate.state=state;
				i2c_out_buf.PFstate.newrx=pf_newrx;
				i2c_out_msg=&i2c_out_buf.u8[0];
				i2c_out_msg_len=3;
				break;
			case 0x41:
				i2c_proc_cnt=0;
				//i2c_out_msg_len=0;
				//i2c_out_msg=&i2c_dbg_buf[0];
			case 0x42:
			case 0x43:
			case 0x44:
			case 0x45:
			case 0x46:
			case 0x47:
			case 0x48:
			case 0x49:
			case 0x4A:
			case 0x4B:
			case 0x4C:
			case 0x4D:
			case 0x4E:
			case 0x4F:
			case 0x50:
				switch (i2c_in_msg_len)
				{
					case 1: // just the "address" byte, host wants to read
						i2c_out_buf=i2c_out_msgA[i2c_in_buf[0]-0x41]; // copy, in case another arrives
						i2c_out_msg=&i2c_out_buf.u8[0];
						i2c_out_msg_len=3;
						break;
					case 4: // host wants to send
						i2c_snd_msgA[i2c_in_buf[0]-0x41]=*((PFmsg*) &i2c_in_buf[1]);
						break;
					default:
						break;
						
				}						
				break;
			
			
			default:
				break;
		}
	//i2c_in_buf[0]=0;
	i2c_processed = 1;
	i2c_new_input=0;
}


// Write sequence 0x60 -> 0x80 *N -> 0x88
// Read sequence  0xA8 -> 0xB8 *N -> 0xC0 or 0xC8 
ISR(TWI_vect)
{
//	i2c_dbg_buf[dbg_idx]=TWSR;
//	if (dbg_idx<I2C_MAX_MSG-1)
//		dbg_idx++;

	switch (TWSR & 0xF8) // map out prescaler bits (19.6.3)
	{ 

// Write sequence start
		case 0x60: // received own slave address, master will write
//			dbg_idx=0;
//			i2c_dbg_buf[dbg_idx++]=TWSR;
	
			i2c_in_msg_len = 0;
			i2c_new_input=1;
			i2c_processed=0;
			TWCR |= _BV(TWEA);
			break;
			
		case 0x80: // addressed as slave, data byte received, returned ACK
			i2c_in_buf[ i2c_in_msg_len ] = TWDR;
			if (i2c_in_msg_len > (I2C_MAX_MSG-3))
				TWCR &= ~_BV(TWEA); // next one must be the last one, no more room!
			else
				TWCR |= _BV(TWEA); // can receive more after next
			i2c_in_msg_len++;
			break;
			
		case 0x88: // addressed as slave, data byte received, returned NACK
			i2c_in_buf[ i2c_in_msg_len ] = TWDR;
			TWCR |= _BV(TWEA); // back to idle slave mode
			// i2c_process_in_msg();
			// write from master ended, apart from stop bit
			break;

// Either sequence end			
		case 0xA0: // stop or repeated start while addressed as slave
			if (i2c_new_input && i2c_in_msg_len>0)
				i2c_process_in_msg();
			TWCR |= _BV(TWEA); // back to idle slave mode
			// transaction ended
			break;

// Read sequence start
		case 0xA8: // received own slave address, master will read
/*
			if (pf_cmd_recv_valid)
			{
				i2c_out_msg[0] = (pf_cmd_recv & 0xff);
				i2c_out_msg[1] = (pf_cmd_recv >> 8);
				i2c_out_msg[2] = pf_cmd_cnt;
				i2c_out_msg_len = 3;
				pf_cmd_recv_valid = 0; // don't report same cmd twice
			}
			else
			{
				i2c_out_msg[0] = 0;
				i2c_out_msg[1] = 0;
				i2c_out_msg[2] = 0;
				i2c_out_msg_len = 3;
			}
*/
			// ...and drop through to transmit first byte!
		case 0xB8: // byte sent, master responded with ACK		
			if (i2c_out_msg_len == 0) 			 
				TWDR = 0xA5; // error condition, send a funny pattern
			else 
				 //{ i2c_out_msg_len--;TWDR =i2c_proc_cnt++;}
				TWDR=i2c_out_msg[--i2c_out_msg_len]; // works better than the original!
			
			if (i2c_out_msg_len == 0) 
				TWCR &= ~_BV(TWEA); // last byte; expect NACK
			else 
				TWCR |= _BV(TWEA); // more to send; expect ACK
			break;
		
			//TWDR=*i2c_out_msg++;	
			//TWCR |= _BV(TWEA); // more to send; expect ACK
			//break;
		case 0xC0: // data byte transmitted, master responded with NACK
		case 0xC8: // LAST byte transmitted (expected NACK), master responded with ACK
			TWCR |= _BV(TWEA); // just switch to idle slave mode
			break;

// Error conditions
		case 0x00: // bus error
			i2c_proc_cnt+=10;
			i2c_new_input=0;
			TWCR |= _BV(TWSTO) | _BV(TWEA); // recovery process - see 19.8.5 in data sheet
			break;
			
		default:
			// all other case should not occur in this system; error
			break;
	}
	//TWCR |= _BV(TWINT);
}

void i2c_slave_init()
{
	uint8_t i;


	uart_puts("i2c slave init\r\n");

	cli();

	for (i=0;i<4;i++)
	{
		i2c_out_msgA[i].msg.msg=i*0xB0A;
		i2c_out_msgA[i].msg.cnt=i;
	}

	TWAR = (SLAVE_ADDRESS << 1);
	TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA); // enable module, enable ints, enable ack

	sei();
}

/*
 * Automatic configuration of the I2C pull-ups
 */

void i2c_pullup_config()
{
	char		 buffer[7],txb;
	uint16_t result,prev_result=0xFFFF;

	PRR &= ~_BV(PRADC);	 // make sure that the ADC is not shut down
	ADCSRA =	 _BV(ADEN)	// enable the ADC
				| _BV(ADPS0) | _BV(ADPS1) | _BV(ADPS2); // prescale by 128
					 // to get ADC clock of 50 to 200kHz (here from the 14.7MHz clock).

	ADMUX =	 _BV(REFS0) // use AVcc as a reference
			| _BV(MUX0) | _BV(MUX2); // select ADC5=SCL

	ADCSRA |= _BV(ADSC); // start a conversion, single-conversion mode
	while (ADCSRA & _BV(ADSC)); // wait for completion

	/*
	DDRC |= _BV(DDC3);	// we use PC3 as the pull-up point (not VDD)
	DDRC &= ~_BV(DDC2); // PC2 also has an 82kOhm to SCL, put into high-z
	PORTC |= _BV(PORTC3); // pull up

	ADCSRA |= _BV(ADSC); // do it again, we discard the first conversion
	while (ADCSRA & _BV(ADSC)); // wait for completion

	result = ADC & 0x3FF;
	uart_puts_P("scl voltage with pullup = ");
	itoa(result, buffer, 10);
	uart_puts(buffer);
	uart_puts_P("\r\n");

	DDRC	|= _BV(DDC3) | _BV(DDC2); // PC2 also has an 82kOhm to SCL
	PORTC |= _BV(PORTC2);	// pull up (simulate a pullup on another device)
	PORTC &= ~_BV(PORTC3); // pull down here

	ADCSRA |= _BV(ADSC); // do it again, we discard the first conversion
	while (ADCSRA & _BV(ADSC)); // wait for completion

	result = ADC & 0x3FF;
	uart_puts_P("scl voltage with pulldown+pulldown = ");
	itoa(result, buffer, 10);
	uart_puts(buffer);
	uart_puts_P("\r\n");

	*/

	DDRC |= _BV(DDC3);	// we use PC3 as the pull-up point (not VDD)
	DDRC &= ~_BV(DDC2); // PC2 also has an 82kOhm to SDA, put into high-z
	PORTC &= ~_BV(PORTC3); // pull down

	while (1)
	{
		ADCSRA |= _BV(ADSC); // do it again, we discard the first conversion
		while (ADCSRA & _BV(ADSC)); // wait for completion

		result = ADC & 0x3FF;

		if ((result^prev_result)<8)
			break;
		prev_result=result;

		uart_puts_P("scl voltage with pulldown = ");
		itoa(result, buffer, 10);
		uart_puts(buffer);
		uart_puts_P("\r\n");

		txb=1;
		for (;txb;)
			uart_status(&txb,NULL);
	
	}

	if (result > 128) 
	{
		// we don't need the pullups, something else is pulling it up
		uart_puts_P("disabling local pullups on i2c lines\r\n");
		DDRC &= ~_BV(DDC3); // put PC3 ...
		DDRC &= ~_BV(DDC2); // ...and PC2 into high-z
	}
	else
	{
		// enable pullups
		uart_puts_P("enabling local pullups on i2c lines\r\n");
		DDRC |= _BV(DDC3);
		PORTC |= _BV(PORTC3); // pull up
		DDRC |= _BV(DDC2);
		PORTC |= _BV(PORTC2); // pull up
	}

	ADCSRA = 0; // disable the ADC (ADEN is set to zero here)
	PRR |= _BV(PRADC);	// shut down the ADC

}
